import java.util.Date;
import java.io.PrintStream;

//AI checkers player, built to lose
public class Machine implements Player
{
   //outputs the prunes of Alpha-Beta
   private static final PrintStream out = System.out;

   //Max time to spend searching
   
   
   //Max levels of tree to search
   private int moveLimit;

   private boolean isRed;
   
   private int shallowPrunes = 0;
   private int deepPrunes = 0;

   //Heuristics for different pieces and places on the board
   private static final int PLAYER_PIECE = -20;
   private static final int PLAYER_KING = -40;
   private static final int OPPONENT_PIECE = 10;
   private static final int OPPONENT_KING = 20;
   
   private static final int PLAYER_SQUARES[] = {  /* Back row */ 0,-5,-5,-5,-5,
                                                  /* Row 1    */ -2,-1,-1,-1,
                                                  /* Row 2    */ -2,-2,-2,-3,
                                                  /* Row 3    */ -4,-3,-3,-3,
                                                  /* Row 4    */ -4,-4,-4,-5,
                                                  /* Row 5    */ -6,-5,-5,-5,
                                                  /* Row 6    */ -6,-6,-6,-7,
                                                  /* King row */ -20,-20,-20,-20};
                                                   
   private static final int OPPONENT_SQUARES[] = {/* Back row */ 0,20,20,20,20,
                                                  /* Row 1    */ 7,6,6,6,
                                                  /* Row 2    */ 5,5,5,6,
                                                  /* Row 3    */ 5,4,4,4,
                                                  /* Row 4    */ 3,3,3,4,
                                                  /* Row 5    */ 3,2,2,2,
                                                  /* Row 6    */ 1,1,1,2,
                                                  /* King row */ 5,5,5,5};

   //overloaded constuctor
   public Machine(int move)
   {
      moveLimit = move;
      
   }

   //default constructor


   //representation of the AI Player
   public String getName()
   {
      String result = new String("Alpha-Beta (" + String.valueOf(moveLimit));

         return result + "mv)";
     
         
   }

  //finds heuristic of all board spaces, returns it as int
   private int getHValue(Position position)
   {
      int hValue = 0;

      for (int count = 1; count <= 32; count++)
      {
         if (!position.isEmpty(count))
         {
            if ((position.isBlack(count) && !isRed) || (!position.isBlack(count) && isRed))
            {
               if (!position.isPiece(count)) 
                 {hValue += PLAYER_KING;}
               else
               {
                  hValue += PLAYER_PIECE;
                  if (isRed)
                    {hValue += -OPPONENT_SQUARES[count];}
                  else 
                    {hValue += PLAYER_SQUARES[count];}
               }
            }
            else
            {
               if (!position.isPiece(count)) {hValue += OPPONENT_KING;}
               
               else
               {
                  hValue += OPPONENT_PIECE;
                  if (isRed) 
                    {hValue += -PLAYER_SQUARES[count];}
                  else 
                    {hValue += OPPONENT_SQUARES[count];}
               }
            }   
         }
      }

      return hValue;
   }

//returns next best move, where the AI will move
   public String getMove(Position position)
   {
      deepPrunes = 0;
      shallowPrunes = 0;
      
      
      isRed = position.isRedsMove();

      if (position.getLegalMoves().size() > 1)
      {
         int bestMove = 0;
         int bestMoveValue = -10000;

         for (int count = 0; count < position.getLegalMoves().size(); count++)
         {
            int nextMoveValue = Maximum(1,position.makeMove((String)position.getLegalMoves().get(count)),-10000,10000);
            if (nextMoveValue > bestMoveValue)
            {
                bestMove = count;
                bestMoveValue = nextMoveValue;
            }
         }

         

         

         out.println("Total branches pruned: " + shallowPrunes);
         out.println("   Total twigs pruned: " + deepPrunes + "\n");
         return (String)position.getLegalMoves().get(bestMove);
      }

      else
        {return (String)position.getLegalMoves().get(0);}
   }


//recursive minimax algorithm: Max sector
   private int Maximum ( int depth, Position position, int alpha, int beta)
   {
      if (depth == moveLimit)
        {return getHValue(position);}
      else
      {
         for (int count = 0; count < position.getLegalMoves().size(); count++)
         {
            
            int nextMoveValue = Minimum(depth + 1, 
                            position.makeMove((String)position.getLegalMoves().get(count)),
                                alpha, beta);
            
            if (nextMoveValue >= alpha)
               {alpha = nextMoveValue;}

            if (alpha >= beta)
            {
               if (depth > moveLimit / 2) 
                {deepPrunes++;}
               else
                {shallowPrunes++;}
                
               return beta;
            }
         }
         return alpha;
      }
   }

//recursive minimax algorithm: Min sector
   private int Minimum (int depth, Position position, int alpha, int beta)
   {
      if (depth == moveLimit)
        {return getHValue(position);}
      else
      {
         for (int count = 0; count < position.getLegalMoves().size(); count++)
         {
           

            int nextMoveValue = Maximum(depth + 1, 
                            position.makeMove((String)position.getLegalMoves().get(count)),
                                                alpha, beta);
            if (nextMoveValue < beta)
               {beta = nextMoveValue;}

            if (beta <= alpha)
            {
               if (depth > moveLimit / 2) 
                 {deepPrunes++;}
               else
                 {shallowPrunes++;} 
             
             return alpha;
            }
         }
         return beta;
      }
   }

}
