/*
 * Created on Mar 5, 2004
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */

 

import java.util.Vector;


public class Evaluation {
	private String s;
	private boolean color;
	private char testColor;
	private char testKing;
	
	private int checkers;
	private int kings;
	private int mobility;
	private int exch;
	private int kCent;
	private int frontRank;
	private int guard;
	
	//constructors
	public Evaluation()
	{
	}

	public Evaluation(byte side)
	{
		color = Const.booleanPlayer(side);
	}
	
	public Evaluation(int checkers, int kings, int mobility, int exch, int kCent, int frontRank, int guard)
	{
		this.checkers = checkers;
		this.kings = kings;
		this.mobility = mobility;
		this.exch = exch;
		this.kCent = kCent;
		this.frontRank = frontRank;
		this.guard = guard;		
	}
	
	public void evaluate(Node currentNode) {
		State currentState = currentNode.getState();
		s=currentState.getBoard();
		
		if(color==true){
			testColor = 'r';
			testKing = 'R';
		} else {
			testColor = 'w';
			testKing = 'W';
		}
		
		checkersAndKings();	// checkers and kings
		exch();	// also takes care of mobility
		kCent();
		frontRank();
		guard();
	}

	/*
	 * First and Second Heuristic
	 */
	private void checkersAndKings(){

		checkers=0;
		kings=0;
		for(int i=0;i<s.length();i++){
			if(s.charAt(i)==testColor){
				checkers++;
			}else if(s.charAt(i)==testKing){
				checkers++;
				kings++;
			}
		}
	}
	
	/*
	 * Fourth Heuristic
	 * 
	 */
	public void exch(){
		exch=0;
		Vector S = new Vector();	// 
		Moves C = new Moves();	//
		State B;
		
		C.allPossibleMoves(new Node(new State(s), " "), S , Const.bytePlayer(color));	// get all possible moves for this color
		mobility = S.size();			// set mobility
		while(S.size()>0){								// for all possible hero moves
			Node grab = (Node) S.remove(0);
			B = grab.getState();
			if(quiescent(B,!color)){	// see if enemy is forced to jump
				exch++;
			}
		}
	}

	/*
	 * Fifth Heuristic
	 */
	private void kCent(){
		kCent=0;
		int n=8;
		int m=4;
//		int n=(int)Math.sqrt(2*s.length());
//		int m=n/2;
		for(int i=(n/4);i<(n*3/4);i++){
			for(int j=(m/4);j<(m*3/4);j++){
				if(s.charAt((i*m)+j)==testKing){
					kCent++;
				}
			}
		}
	}
	
	/*
	 * Sixth Heuristic
	 */
	private void frontRank(){
		double value=0;
		if(color){	// red's move
			for(int i=1;i<=5;i++){
				for(int j=0;j<4;j++){
					if((s.charAt((i*4)+j))==testColor){
						int temp =i-2; //(i+1+1-(boardsize/2))^2
						value+=(temp)*temp;	// (i+2-4)^2
					}
				}
				
			}
		} else {	// white's move
			for(int i=4;i<8;i++){
				for(int j=0;j<4;j++){
					if((s.charAt((i*4)+j))==testColor){
						int temp = 5-i; //((boardsize/2)+2-(i+1))^2
						value+=(temp)*temp;		// (8/4)+1-i
					}								
				}
			}
		}
		frontRank = (int)value;
	}
	
	
	/*
	 * Seventh Heuristic
	 */
	private void guard(){
		int tmp=0;
		if(color){
			for(int i=0;i<4;i++){
				if(s.charAt(i)==testColor||s.charAt(i)==testKing){
					tmp++;
				}
			}
		}else{
			for(int i=28;i<32;i++){
				if(s.charAt(i)==testColor||s.charAt(i)==testKing){
					tmp++;
				}
			}
		}
		if(tmp<4){
			guard = 0;
		} else {
			guard = 1;
		}
	}

	public int getCheckers(){
		return checkers;
	}
	public int getKings(){
		return kings;
	}
	public int getMobility(){
		return mobility;
	}
	public int getExch(){
		return exch;
	}
	
	public int getkCent(){
		return kCent;
	}
	public int getFrontRank(){
		return frontRank;
	}
	public int getGuard(){
		return guard;
	}
	

	public boolean quiescent(State currentState, boolean ncolor){
		
		/** Initialize variables */
		
//		int height = (int)Math.sqrt(2*currentBoard.getSize());
//		int width = height/2;
		int height=8;
		int width=4;
		
		char king;
		char notKing;
		char jumpK;
		char jumpNK;

		
		// calculate offsets to adjacent spots
		//int[] adjKing=new int[4];
		int adjL_NE = -(width-1);
		int adjL_NW = -width;
		int adjL_SW = width;
		int adjL_SE = width+1;
		
		int adjR_NE=-width;
		int adjR_NW=-(width+1);
		int adjR_SW=width-1;
		int adjR_SE=width;
		
		// NE, NW, SW, SE
		int[] jumpKing = {(adjL_NE+adjR_NE), (adjL_NW+adjR_NW), (adjL_SW+adjR_SW), (adjL_SE+adjR_SE)}; 
		int[] jump=new int[2];

		if(ncolor){
			king='R';
			notKing='r';
			jump[0]=jumpKing[2];	// southWest jump
			jump[1]=jumpKing[3]; // southEast jump
		} else {
			king='W';
			notKing='w';
			jump[0]=jumpKing[0];	// northEast jump
			jump[1]=jumpKing[1]; // northWest jump
		}
		
		/** Gather current pieces */
	
		String state = currentState.getBoard();
		
		// get locations of all movable pieces
		Vector NotKings = new Vector();
		Vector Kings = new Vector();
		for(int i=0;i<state.length();i++){
			if(state.charAt(i)==king) {
				Kings.add(new Integer(i));
			} else if(state.charAt(i)==notKing){
				NotKings.add(new Integer(i));
			}
		}
		
		// Check all adjacent spots for a king
		for(int i=0;i<Kings.size();i++){
			int tmp = ((Integer) Kings.remove(0)).intValue();
			if(tmp%height<width){	// adjacents are left staggered
				// ne,nw,sw,se
				int[] adjKing = {adjL_NE,adjL_NW,adjL_SW,adjL_SE};
				if(move(state, adjKing, jumpKing, tmp, ncolor)==true){ return true; }
			} else {		// adjacents are right staggered
				int[] adjKing = {adjR_NE,adjR_NW,adjR_SW,adjR_SE};
				if(move(state, adjKing, jumpKing, tmp, ncolor)==true){ return true;	}
			}
		}
		
		// check appropriate adjacent spots for not king
		for(int i=0;i<NotKings.size();i++){
			// check front adjacent spots
			int tmp = ((Integer) NotKings.remove(0)).intValue();
			
			if(tmp%height<width){// left staggered
				if(ncolor){ // only south, left staggered
					// SW, SE
					int[] adj = {adjL_SW,adjL_SE};
					if(move(state, adj, jump, tmp, ncolor)==true){	return true;	}
				}else{		// only north, left staggered
					// NE,NW
					int[] adj = {adjL_NE,adjL_NW};
					if(move(state, adj, jump, tmp, ncolor)==true){	return true;	}
				}
			} else {		//  right staggered
				if(ncolor){	// only south, right staggered
					// SW,SE
					int[] adj = {adjR_SW,adjR_SE};
					if(move(state, adj, jump, tmp, ncolor)==true){	return true;	}
				}else{	// only north, right staggered
					// NE,NW
					int[] adj = {adjR_NE,adjR_NW};
					if(move(state, adj, jump, tmp, ncolor)==true){	return true;	}
				}
			}
		}
		return false;
	}
	
	private boolean move(String state, int[] adj, int[] jump, int pos, boolean ncolor){
		
		/** Initialize variables */
		char king;
		char notKing;
		char jumpK;
		char jumpNK;
		int height;
		int width;
		
//		height = (int)Math.sqrt(2*state.length());
//		width = height/2;
		height=8;
		width=4;
		
		if(ncolor){
			king='R';
			notKing='r';
			jumpK='W';
			jumpNK='w';
		} else {
			king='W';
			notKing='w';
			jumpK='R';
			jumpNK='r';

		}
		
		char testVal;
		int newLoc;
		
		for(int i=0; i<adj.length; i++){	// for each adjacent
			newLoc = pos+adj[i];			// adjacent space
			if(newLoc<=31 && newLoc>=0){	// if on board
				testVal = state.charAt(newLoc);	// check adjacent
				newLoc=pos+jump[i];			// calculate jump
				if((testVal==jumpK || testVal==jumpNK) && newLoc<=31 && newLoc>=0){
					testVal = state.charAt(newLoc);
					// Enemy is Adjacent, so check for jumps
					if(state.charAt(newLoc)=='b'){		// jump is open
									return true;
					}
				}
			}
		}
		return false;
	}
	
	void deltaEvaluation(Evaluation plus, Evaluation minus, Evaluation result)
	{
		result.checkers = plus.checkers - minus.checkers;
		result.kings = plus.kings - minus.kings;
		result.mobility = plus.mobility - minus.mobility;
		result.exch = plus.exch - minus.exch;
		result.kCent = plus.kCent - minus.kCent;
		result.frontRank = plus.frontRank - minus.frontRank;
		result.guard = plus.guard - minus.guard;
	}

	public void print(){
		System.out.println("Checkers: "+checkers);
		System.out.println("Kings: "+kings);
		System.out.println("Mobility: "+mobility);
		System.out.println("Exch: "+exch);
		System.out.println("kCent: "+kCent);
		System.out.println("frontRank: "+frontRank);
		System.out.println("guard: "+guard);
		System.out.println();
	}	
}
