import java.util.*;
/**
 * Implements a Minimax search algorithm with alpha beta pruning
 * 
 * @author Cheryl Yager
 * @author Umar K. Munroe
 * @version 1.0
 */
public class AlphaBetaPruningSearch
{

	private State newState;

	private MinimaxNode[] rootSuccessors = new MinimaxNode[0];

	public AlphaBetaPruningSearch(State s)
	{
		long startTime = System.currentTimeMillis();
		MinimaxNode tree;
		MinimaxNode.setStartTime(startTime);
		tree = new MinimaxNode(s);
		newState = alphaBetaSearch(tree);
		System.out.println("Elapsed Time: "
				+ ((System.currentTimeMillis() - startTime) / 1000.0) + " ms");
	}

	public State getBestMove()
	{
		return newState;
	}
	
	
	private State alphaBetaSearch(MinimaxNode node)
	{
		int utilityValue = maxValue(node);
		System.out.println("utilityValue " + utilityValue);
		MinimaxNode[] successors = rootSuccessors;
		State answer = node.getState();
		TreeSet list = new TreeSet();
		if (successors.length == 1)
		{
			answer = successors[0].getState();
		}
		else if (successors.length == 0)
		{
		}
		else
		{
			for (int i = 0; i < successors.length; i++)
			{
				if (successors[i].getAlpha() == utilityValue
						|| successors[i].getBeta() == utilityValue)
				{
					list.add(successors[i]);
				}
			}
			answer = ((MinimaxNode) list.first()).getState();
		}
		successors = new MinimaxNode[list.size()];
		list.toArray(successors);
		for (int i = 0; i < successors.length; i++)
		{
		}
		return answer;
	}

	public int maxValue(MinimaxNode node)
	{
		int alpha = node.getAlpha();
		int beta = node.getBeta();
		int utilityValue;
		MinimaxNode[] successors = node.getSuccessors();
		if (successors.length > 0 && successors[0].getDepth() == 1)
		{
			rootSuccessors = successors;
		}
		if (node.isUtilitySet())
		{
			utilityValue = node.getUtility();
		}
		else
		{
			utilityValue = Integer.MIN_VALUE;
			node.setUtility(utilityValue);
			boolean breakLoop = false;
			for (int i = 0; i < successors.length && breakLoop == false; i++)
			{
				utilityValue = Math.max(utilityValue, minValue(successors[i]));
				node.setUtility(utilityValue);
				if (utilityValue >= beta)
				{
					breakLoop = true;
				}
				else
				{
					alpha = Math.max(alpha, utilityValue);
					node.setAlpha(alpha);
				}
			}
		}
		return utilityValue;
	}

	public int minValue(MinimaxNode node)
	{
		int alpha = node.getAlpha();
		int beta = node.getBeta();
		int utilityValue;
		MinimaxNode[] successors = node.getSuccessors();
		if (node.isUtilitySet())
		{
			utilityValue = node.getUtility();
		}
		else
		{
			utilityValue = Integer.MAX_VALUE;
			node.setUtility(utilityValue);
			successors = node.getSuccessors();
			boolean breakLoop = false;
			for (int i = 0; i < successors.length && breakLoop == false; i++)
			{
				utilityValue = Math.min(utilityValue, maxValue(successors[i]));
				node.setUtility(utilityValue);
				if (utilityValue <= alpha)
				{
					breakLoop = true;
				}
				else
				{
					beta = Math.min(beta, utilityValue);
					node.setBeta(beta);
				}
			}
		}
		return utilityValue;
	}
}
