import java.io.*;
/**
 * Instantiates a one player vs. AI agent 1) Allows for the selection of what
 * color (red or white) the user wants to be whether or not the user wants to
 * go first. 2) Leverages a Minimax search with alpha-beta pruning for the
 * computer agent to play against the user. 3) Provides the user the ability to
 * confirm each move. 4) Provides undo of a single move if available
 * 
 * @author Cheryl Yager
 * @author Umar K. Munroe
 * @version 1.0
 */
public class Checkers
{

	/**
	 * Constructor for objects of class Checkers
	 */
	public Checkers()
	{
	}

	/**
	 * Plays a game of checkers
	 * 
	 * @param agentIsRed true if the agent is red, false if the agent is white
	 * @param redIsFirst true if the red goes first, false if wite goes first
	 */
	public void playGame(boolean agentIsRed, boolean redIsFirst)
	{
		//intializes the board indicating who goes first
		State currentState = new State(redIsFirst);
		String currentBoard = "z";
		String previousBoard = currentState.getBoard();
		AlphaBetaPruningSearch search;
		int[] opponentsMove;
		boolean areYouSure;
		boolean agentIsFirst = (redIsFirst && agentIsRed)
				|| (!redIsFirst && !agentIsRed);
		boolean gameOver = false;
		boolean gameAbort = false;
		boolean agentWins = false;
		char select;
		boolean startAgent = false;
		boolean startOpponent = false;
		int numAgentMoves = 0;
		int numOpponentMoves = 0;
		State previousAgentState = null;
		String input;
		BufferedReader in;
		State.setAgentRed(agentIsRed);
		currentState.displayBoard();
		try
		{
			Runtime
					.getRuntime()
					.exec(
							"\"C:/Program Files/Windows Media Player/wmplayer.exe\" /play /close \"C:/Documents and Settings/cslabuser/Desktop/start.wav\"");
		}
		catch (Exception e)
		{
		}
		if (agentIsFirst)
		{
			startAgent = true;
		}
		else
		{
			startOpponent = true;
		}
		try
		{
			//while the game has not ended
			while (gameOver == false && gameAbort == false)
			{
				if (startAgent && gameOver == false)
				{
					startOpponent = true;
					search = new AlphaBetaPruningSearch(currentState);
					//make agent's move
					if (currentState.myNumPieces() > 0
							&& currentState.otherNumPieces() > 0)
					{
						currentState = search.getBestMove();
						numAgentMoves++;
					}
					currentBoard = currentState.getBoard();
					//if the agent's move left the board as it was before
					//it couldn't move, so it wins
					if (currentBoard.compareTo(previousBoard) == 0)
					{
						gameOver = true;
						agentWins = true;
					}
					//make the agent's move the previous board
					previousBoard = currentBoard;
					//display Move
					currentState.displayBoard();
				}
				if (startOpponent && gameOver == false)
				{
					startAgent = true;
					if (currentState.myNumPieces() > 0
							&& currentState.otherNumPieces() > 0)
					{
						do
						{
							//get opponents move
							opponentsMove = Move.getMove();
							select = getYesOrNoInput("Are you sure?");
							if (opponentsMove == null)
							{
								//performUndo();
								if (previousAgentState == null)
								{
									System.out
											.println("Cannot UNDO at this point!");
								}
								else
								{
									System.out
											.println("*********************************************");
									System.out
											.println("*******Reverted board before last move*******");
									System.out
											.println("*********************************************");
									System.out.println();
									currentState = previousAgentState;
									previousAgentState = null;
									System.gc();
									numAgentMoves--;
									numOpponentMoves--;
									currentState.displayBoard();
								}
								select = 'n';
							}
							if (select == 'y')
							{
								areYouSure = true;
							}
							else if (select == 'n')
							{
								areYouSure = false;
							}
							else
							{
								gameAbort = true;
								areYouSure = true;
							}
						}
						while (currentState.isMoveValid(opponentsMove) == false
								|| areYouSure == false);
						previousAgentState = new State(currentState.getBoard(),
								currentState.getNumJumps(), currentState
										.getMadeMoves());
						previousAgentState.setTurn(currentState.isRedTurn());
						currentState = currentState
								.makeOpponentMove(opponentsMove);
						numOpponentMoves++;
					}
					currentBoard = currentState.getBoard();
					//if the opponent's move left the board as it was
					// before it couldn't move, so it wins
					if (currentBoard.compareTo(previousBoard) == 0)
					{
						//System.out.println("in here");
						gameOver = true;
						agentWins = false;
					}
					else
					{
						currentState.displayBoard();
					}
					previousBoard = currentBoard;
				}
			}
			if (agentWins)
			{
				System.out.println("Agent Wins!");
				try
				{
					Runtime
							.getRuntime()
							.exec(
									"\"C:/Program Files/Windows Media Player/wmplayer.exe\" \"C:/Documents and Settings/cslabuser/Desktop/win.wav\"");
				}
				catch (Exception e)
				{
				}
			}
			else
			{
				System.out.println("You Win!....lucky");
				try
				{
					Runtime
							.getRuntime()
							.exec(
									"\"C:/Program Files/Windows Media Player/wmplayer.exe\" \"C:/Documents and Settings/cslabuser/Desktop/lose.wav\"");
				}
				catch (Exception e)
				{
				}
			}
			if (gameAbort == true)
			{
				System.out.println("Game aborted!");
			}
			System.out.println("Total Moves:  Agent = " + numAgentMoves
					+ "\tOpponent = " + numOpponentMoves);
		}
		catch (IOException e)
		{
			e.printStackTrace();
			throw new RuntimeException();
		}
	}

	public static void main(String[] args) throws Exception
	{
		InputStreamReader isr;
		StreamTokenizer st;
		String input;
		int choice;
		boolean redFirst = false;
		boolean redAgent = false;
		boolean proceed = true;
		Checkers theGame = new Checkers();
		System.out
				.println("Welcome to checkers against the SquarePants that needs no introduction!\n\n");
		System.out
				.println("What color would you like to be? (1 = Red, 2 = White):");
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		input = in.readLine();//nextToken();
		if (input != null)
		{
			try
			{
				choice = Integer.parseInt(input);
			}
			catch (NumberFormatException e)
			{
				choice = 0;
			}
			switch (choice)
			{
				case 1 :
					redAgent = false;
					break;
				case 2 :
					redAgent = true;
					break;
				default :
					System.out
							.println("Invalid input!  C'mon, this isn't exactly rocket science.");
					System.out.println("You've made me sad.  I'm aborting.");
					proceed = false;
					break;
			}
			if (proceed)
			{
				char select = getYesOrNoInput("Do you want to go first?\t");
				switch (select)
				{
					case 'y' :
						if (redAgent)
						{
							redFirst = false;
						}
						else
						{
							redFirst = true;
						}
						theGame.playGame(redAgent, redFirst);
						break;
					case 'n' :
						if (redAgent)
						{
							redFirst = true;
						}
						else
						{
							redFirst = false;
						}
						theGame.playGame(redAgent, redFirst);
						break;
					default :
						System.out
								.println("You've made me sad.  I'm aborting.");
						break;
				}
			}
		}
		System.out.println("\nDone!");
	}

	/**
	 * Gets a y or n response from the user for the input question from the
	 * user at the command line .
	 */
	public static char getYesOrNoInput(String question) throws IOException
	{
		BufferedReader in;
		boolean validInput = false;
		String input = null;
		char select = 'x';
		while (validInput == false)
		{
			System.out.println(question);
			System.out.println("(y = YES,  n = NO, x = ABORT)?\t");
			in = new BufferedReader(new InputStreamReader(System.in));
			input = in.readLine();
			input = input.trim().toLowerCase();
			if (input.length() == 1)
			{
				select = input.charAt(0);
				if (select == 'y' || select == 'n' || select == 'x')
				{
					validInput = true;
				}
			}
			if (validInput == false)
			{
				System.out.println("Invalid input!  C'mon, this isn't "
						+ "exactly rocket science.");
			}
		}
		return select;
	}
}
