/**
 * Implements the minimax algorithm 1) Sets the max depth for searching to 8 to
 * maximize stability. (Going deeper than 8 ply occasionally produced a memory
 * error due to our state representation) 2) Allows 20 sec. for the search
 * 
 * @author Umar Munroe
 * @author Cheryl Yager
 * @version 1.0
 */
public class MinimaxNode implements Comparable
{

	//the state
	private State state;

	//alpha-beta pruning alpha value
	private int alpha = Integer.MIN_VALUE;

	//alpha-beta pruning beta value
	private int beta = Integer.MAX_VALUE;

	private int depth;

	private static long startTime;

	private static final int MAX_DEPTH = 8;

	private static final long MAX_TIME = 20 * 1000;

	static int nodeCount = 0;

	private int utility;

	private boolean utilitySet = false;

	public MinimaxNode(State s)
	{
		this(s, -1);
	}

	/**
	 * Constructor for objects of class MinimaxNode
	 */
	public MinimaxNode(State s, int parentDepth)
	{
		state = s;
		depth = parentDepth + 1;
		alpha = Integer.MIN_VALUE;
		beta = Integer.MAX_VALUE;
		if (depth == MAX_DEPTH)
		{
			//System.out.println("MAX_DEPTH = "+ depth);
		}
		nodeCount++;
	}

	protected static void setStartTime(long theStartTime)
	{
		startTime = theStartTime;
	}

	public int getBeta()
	{
		return beta;
	}

	public int getAlpha()
	{
		return alpha;
	}

	public MinimaxNode[] getSuccessors()
	{
		MinimaxNode[] successors;
		State[] validSuccessors = new State[0];
		long currentTime = System.currentTimeMillis();
		if (currentTime - startTime < MAX_TIME && depth < MAX_DEPTH)
		{
			validSuccessors = state.getSuccessors();
		}
		successors = new MinimaxNode[validSuccessors.length];
		for (int i = 0; i < successors.length; i++)
		{
			successors[i] = new MinimaxNode(validSuccessors[i], depth);
		}
		if (validSuccessors.length == 0)
		{
			setUtility();
		}
		return successors;
	}

	public State getState()
	{
		return state;
	}

	public void setBeta(int value)
	{
		beta = value;
	}

	public void setAlpha(int value)
	{
		alpha = value;
	}

	public boolean isUtilitySet()
	{
		return utilitySet;
	}

	public int getUtility()
	{
		if (utilitySet == false)
		{
			throw new UtilityNotSetException();
		}
		return utility;
	}

	public void setUtility()
	{
		utility = state.calcUtility();
		alpha = utility;
		beta = utility;
		utilitySet = true;
	}

	public void setUtility(int val)
	{
		utility = val;
		utilitySet = true;
	}

	public int getDepth()
	{
		return depth;
	}

	public void display()
	{
		System.out.print("alpha = " + alpha);
		System.out.print("\tbeta = " + beta);
		System.out.print("\tisUtilitySet() = " + isUtilitySet());
		if (isUtilitySet())
		{
			System.out.print("\tutility = " + getUtility());
			
		}
		System.out.println();
	}

	public MinimaxNode getParent()
	{
		return null;
	}

	public int compareTo(Object obj)
	{
		return ((MinimaxNode) obj).getUtility() - getUtility();
	}
}
