/**
 * Union College CSc-244 AI Project 3 - succStates >
 * succStates is a class that when instantiated the findSuccs method given a
 * checker board and a color that indicates who the agent is () = RED, 1 = 
 * WHITE), will return all the successor states.
 *
 * @author  TJ Bush & RJ Conti
 * @version 03.11.2004
 */

import java.util.*;     // for Vector

public class succStates
{
    // instance variable(s) ------------------------------------------------
    private char rChecker  = 'r';    // RED checker symbol
    private char rKing     = 'R';    // RED king symbol
    private char wChecker  = 'w';    // WHITE checker symbol
    private char wKing     = 'W';    // WHITE king symbol
    private char canMove   = ' ';    // position checker can move symbol

    // look up table for locations based on row and column for this
    // checker board, 99 is an unused value for a location in this problem
    private int [][] locations = { {99,  0, 99,  1, 99,  2, 99,  3},
                                   { 4, 99,  5, 99,  6, 99,  7, 99},
                                   {99,  8, 99,  9, 99, 10, 99, 11},
                                   {12, 99, 13, 99, 14, 99, 15, 99},
                                   {99, 16, 99, 17, 99, 18, 99, 19},
                                   {20, 99, 21, 99, 22, 99, 23, 99},
                                   {99, 24, 99, 25, 99, 26, 99, 27},
                                   {28, 99, 29, 99, 30, 99, 31, 99}  };    
    //----------------------------------------------------------------------
    
    // constructor method(s) -----------------------------------------------
    // None
    //----------------------------------------------------------------------

    // instance method(s) --------------------------------------------------    
    /**
     * findSuccs()
     * 
     * Generates the successors states and places them in a stack that gets
     * returned to the caller. The method uses recursive calls to determine
     * multiple jump locations.
     * 
     * Note: since the rules of Project 3 include forcing a jump if one
     * exists, the following successors code will not return moves in 
     * instances where a jump is found. I.e., there are a total of four
     * possible moves (in this case moves means moves or jumps) from each
     * location. When an algorithm finds a jump for a particular location,
     * it will disregard any moves from that location and just return the
     * jump(s).
     * 
     * @param       board   -   checker board to find successors states for
     *                          (0 - 31)
     *              color   -   color of side who is looking for successors
     *                          0 = RED, 1 = WHITE
     * @return      A stack of successor states (checker boards)
     */
    public Stack findSuccs(char[] board, int color)
    {
        Stack   successors  = new Stack();  // stores successor nodes
        Vector  locations   = new Vector(); // stores locations of color's 
                                            // pieces
        Vector  diags       = new Vector(); // diagonal locations for a 
                                            // specified location on the 
                                            // checker board

        // First need to determine whose successors we need to find (i.e.,
        // RED's (color = 0) or WHITES's (color = 1) and then find where
        // the respective pieces are on the board provided
        if (color == 0) // looking for RED checker locations
        {
            locations = findLocations(board, 0);  // RED locations
        } // end if looking for RED locations
        
        else            // looking for WHITE checker locations
        {
            locations = findLocations(board, 1); // WHITE locations
        } // end else looking for WHITE locations
            
        // Need to find the successors amongst the array of piece
        // locations found above. The for i loop looks at all color's 
        // pieces on the board.
        for (int i = 0; i < locations.size(); i++)
        {
            boolean canMakeJumps = false;   // flag to identify if jumps
                                            // can be made per iteration of
                                            // location
            
            // the current board position (0 - 31)
            int curPos = ((Integer)locations.get(i) ).intValue();

            // find all the diagonal locations associated with the 
            // specified location on the board (0 - 31)
            diags = findDiags(board, curPos);

            // for each diagonal location
            for (int j = 0; j < diags.size(); j++)
            {
                int diagPos = (((Integer)diags.get(j)).intValue());

                // Find jump successor states.....
                // This will execute only if a jump can be made . This section uses 
                // recursion. canJump != 99 is indicative of a valid jump condition.
                if (canJump(board, curPos, diagPos) != 99)
                {
                    int finish = canJump(board, curPos, diagPos);
                    char[] copy = copyBoard(board);
                    copy = jump(copy, curPos, diagPos, finish); // perform jump
                    aNode succState = new aNode(copy);  // successor
                    succState.addMove(curPos);          // add start point

                    // look for more jumps recursively
                    Stack moreJumps = new Stack();

                    // only look for more jumps if not a crowning event
                    if (crowningEvent(board, curPos, finish) == false)
                    {
                        moreJumps = findMoreJumps(copyBoard(copy), finish);
                    } // end if

                    if (moreJumps.size() == 0)
                    {
                        succState.addMove(finish);      // add finish point
                        succState.incNumJumps();        // increase jump number
                        successors.push(succState);     // push node on stack
                    } // end if no additional jumps
                    
                    // more jumps were found so add them to the stack to return
                    else
                    {
                        int size = moreJumps.size();
                        for (int m = 0; m < size; m++)
                        {
                            aNode another  = succState.copyNode();
                            aNode tempNode = moreJumps.pop();
                            another.setBoard(tempNode.getBoard());
                            for (int k = 0; k < tempNode.getMoves().size(); k++)
                            {
                                another.addMove(((Integer)tempNode.getMoves().get(k)).intValue());
                                another.incNumJumps();
                            } // end for k
                            successors.push(another);
                        } // end for m
                    } // end else more jumps
                // set to flag to true to prevent subsequent calcualtions that will
                // calculate and push moves on the stack of successors
                canMakeJumps = true;
                } // end if can jump 
            } // end for j diagonal location

            // Only execute this is not jumps were previously found for this iteration
            if (canMakeJumps != true)
            {
                // for each diagonal location
                for (int j = 0; j < diags.size(); j++)
                {
                    int diagPos = (((Integer)diags.get(j)).intValue());
    
                    // Find moves successor states.....
                    // Check to see if position is empty, meaning that it is a
                    // potential move location, and if the piece can perform a 
                    // legal move based on its type (color, king, etc.), and if
                    // so, then push the move on the stack of successors
                    if (isValidMove(board, curPos, diagPos) == true)
                    {
                        char[] copy = copyBoard(board);
                        copy = move(copy, curPos, diagPos); // perform move
                        aNode succState = new aNode(copy);  // successor
                        succState.addMove(curPos);          // add start point
                        succState.addMove(diagPos);         // add finish point
                        successors.push(succState);         // push node on stack
                    } // end if empty and valid move
                } // end for j diagonal
            } // end if canMakeJumps != true
        } // end for i locations
        
        // randomly jumbles the order of the successors in the stack prior to 
        // returning it to try and make the overall game the agent plays less
        // predictable
        return successors.jumbleStackRandomly();
    } // end findSuccs()

    /**
     * crowningEvent()
     * 
     * Given a board, the current piece's location, and the location the piece
     * wants to jump to, this method determines if the jump involves a crowning.
     * A crowning will prevent a piece from continuing on a run.
     * 
     * @param       board  - the board
     *              start  - the current postion to test for continuing
     *                      jumps (0 - 31)  
     *              finish - the location the piece is jumping to (0 - 31)
     * @return      true if a crowning event, else false
     */
    private boolean crowningEvent(char[] board, int start, int finish)
    {
        boolean crowning = false;
        if ( (((finish == 0) || (finish == 1) || (finish == 2) || (finish == 3)) 
                   && (board[start] == wChecker)) 
                                            ||
              (((finish == 28) || (finish == 29) || (finish == 30) || (finish == 31)) 
              && (board[start] == rChecker)) 
              )
        {
            crowning = true;
        } // end if crowning
        return crowning;
    } // end crowningEvent()

    /**
     * findMoreJumps()
     * 
     * Given a board, the current position on that board, and a location
     * to look at for a move, this method determines if the piece at the 
     * current position can legally move there. This is a recursive method.
     * 
     * @param       board - the board
     *              start - the current postion to test for continuing
     *                      jumps (0 - 31)  
     * @return      stack of additional jumps
     */
     private Stack findMoreJumps(char[] board, int start)
     {
        Stack jumpStates = new Stack();
        Vector diags     = new Vector();    // diagonal locations for a 
                                            // specified location on the 
                                            // checker board

        
        // find all the diagonal locations associated with the 
        // specified location on the board (0 - 31)
        diags = findDiags(board, start);

        // for each diagonal location
        for (int j = 0; j < diags.size(); j++)
        {
            // Find additional jump successor states.....
            // this will execute only if a jump can be made. This 
            // section uses recursion. canJump != 99 is indicative 
            // of a valid jump condition.
            int diagPos = (((Integer)diags.get(j)).intValue());
            if (canJump(board, start, diagPos) != 99)
            {
                int finish = canJump(board, start, diagPos);
                char[] copy = copyBoard(board);
                copy = jump(copy, start, diagPos, finish); // perform jump
                aNode succState = new aNode(copy);  // successor
                succState.addMove(start);          // add start point
                Stack moreJumps  = new Stack();

                // look for more jumps recursively
                // only look for more jumps if not at the edge of the board
                if (crowningEvent(board, start, finish) == false)
                {
                    moreJumps = findMoreJumps(copyBoard(copy), finish);
                } // end if
                
                if (moreJumps.size() == 0)
                {
                    succState.addMove(finish);      // add finish point
                    succState.incNumJumps();        // increase jump number
                    jumpStates.push(succState);     // push node on stack
                } // end if no additional jumps
                
                else
                {
                    int size = moreJumps.size();
                    for (int m = 0; m < size; m++)
                    {
                        aNode another  = succState.copyNode();
                        aNode tempNode = moreJumps.pop();
                        another.setBoard(tempNode.getBoard());
                        for (int k = 0; k < tempNode.getMoves().size(); k++)
                        {
                            another.addMove(((Integer)tempNode.getMoves().get(k)).intValue());
                            another.incNumJumps();
                        } // end for k
                        jumpStates.push(another);
                    } // end for m
                } // end else
            } // end if can jump
        } // end for j
        return jumpStates;
     } // end findMoreJumps()

    /**
     * isValidMove()
     * 
     * Given a board, the current position on that board, and a location
     * to look at for a move, this method determines if the piece at the 
     * current position can legally move there.
     * 
     * @param       board   - checker board to evaluate
     *              curPos  - the current position on the board (0 - 31)
     *              moveLoc - the location to look at for a potential move
     *                        (0 - 31)
     * @return      true if curPos can move to moveLoc, else false
     */
    private boolean isValidMove(char[] board, int curPos, int moveLoc)
    {
        boolean canMove = false;
        
        // first need to determine if the location to move to is empty
        if (isEmpty(board, moveLoc) == true)
        {
            // check to see if the current piece is a king
            if ( (board[curPos] == rKing) ||
                 (board[curPos] == wKing) )
            {
                canMove = true;
            } // end if king

            // check to see if a RED checker and move is in the legal
            // direction (NORTH to SOUTH)
            else if ( (board[curPos] == rChecker) &&
                      (checkerRow(moveLoc) > checkerRow(curPos)) )
            {
                canMove = true;
            } // end if RED checker

            // check to see if a WHITE checker and move is in the legal
            // direction (SOUTH to NORTH)
            else if ( (board[curPos] == wChecker) &&
                      (checkerRow(moveLoc) < checkerRow(curPos)) )
            {
                canMove = true;
            } // end if WHITE checker
        } // end if moveLoc is empty
        return canMove;
    } // end isValidMove()

    /**
     * findLocations()
     * 
     * Given a board and the color to evaluate, this method will return
     * a vector holding the locations of the specified color's pieces on
     * the board, (0 - 31).
     * 
     * @param       board   -   checker board to evaluate
     *              color   -   color of side to evaluate
     *                          0 = RED, 1 = WHITE
     * @return      An vector holding the location of color's pieces
     *              (0 - 31) on the checker board provided
     */
    private Vector findLocations(char[] board, int color)
    {
        Vector locations = new Vector();
        
        for (int i = 0; i < board.length; i++)
        {
            if (color == 0)         // RED
            {
                if (board[i] == rChecker || board[i] == rKing)
                {
                    locations.addElement(new Integer(i));
                } // end if a RED checker
            } // end if RED side

            else if (color == 1)    // WHITE
            {
                if (board[i] == wChecker || board[i] == wKing)
                {
                    locations.addElement(new Integer(i));
                } // end if a WHITE checker                
            } // end else if WHITE SIDE
        } // end for i
        return locations;
    } // end findLocations()

    /**
     * checkerRow()
     * 
     * Based on the location on the board (0 - 31), this method returns the
     * row number (0 - 7).
     * 
     * @param       index - the location on the board (0 - 31)
     * @return      the row the index value is in (0 - 7)
     */
    private int checkerRow(int index)
    {
        return (index / 4);
    } // end checkerRow()

    /**
     * checkerCol()
     * 
     * Based on the location on the board (0 - 31), this method returns the
     * column number (0 - 7).
     * 
     * @param       index - the location on the board (0 - 31)
     * @return      the column the index value is in (0 - 7)
     */
    private int checkerCol(int index)
    {
        if (checkerRow(index) % 2 == 0)     // even row
        {
            return (2 * (index % 4) + 1);
        } // end if even row

        else                                // odd row
        {
            return (2 * (index % 4));
        } // end else odd row
    } // end checkerCol()

    /**
     * findDiags()
     * 
     * Given a position on a board (0 - 31) and the board this method 
     * finds the diagonal locations. Note that this method uses the 
     * locations look-up table to translate a row and column into a 
     * checker board location 0 - 31. This was done to spare some
     * computation time. For the board location references, NORTH refers
     * to the RED side where the position numbering starts (i.e., 0 , 1,
     * 2, 3, ...).
     * 
     * @param       board - the checker board
     *              index - the location on the board (0 - 31) that the
     *              diagonals need to be caluclated for
     * @return      the diagonal locations fpr the index (0 - 31)
     */
    private Vector findDiags(char[] board, int index)
    {
        Vector diags = new Vector();    // holds the diagonals
        
        int row = checkerRow(index);    // checker's row
        int col = checkerCol(index);    // checker's column

        // find diagonals locations to the NORTH
        if ((row - 1) >= 0)       // row to the N?
        {
            if ((col - 1) >= 0)   // NW
            { 
                diags.addElement(new Integer(locations[(row - 1)][(col - 1)]));
            } // end if NW
            if ((col + 1) <= 7)    // NE
            {
                diags.addElement(new Integer(locations[(row - 1)][(col + 1)]));
            } // end if NE
        } // end row NORTH?

        // find diagonal locations to the SOUTH
        if ((row + 1) <= 7)       // row to the S?
        {
            if ((col - 1) >= 0)   // SW
            {
                diags.addElement(new Integer(locations[(row + 1)][(col - 1)]));
            } // end if SW
            if ((col + 1) <= 7)    // SE
            {
                diags.addElement(new Integer(locations[(row + 1)][(col + 1)]));
            } // end if SE
        } // end row SOUTH?
        return diags;
    } // end findDiags()

    /**
     * copyBoard()
     * 
     * Given a board, this method returns a copy of it.
     * 
     * @param       board - board to copy
     * @return      a copy of board
     */
    private char[] copyBoard(char[] board)
    {
        char[] copy = new char[board.length];
        
        for (int i = 0; i < board.length; i++)
        {
            copy[i] = board[i];
        } // end for i
        return copy;
    } // end copyBoard()

    /**
     * move()
     * 
     * Given a board, and start and finish positions on that board, this
     * method moves the piece at position start to the finish position.
     * This method also performs the crowning of pieces that reach the
     * opposite side of the board.
     *  
     * @param       board  - the board
     *              s      - the starting position for the move (0 - 31)
     *              f      - the ending position for the move (0 - 31)
     * @return      the manipulated board 
     */
    private char[] move(char[] board, int s, int f)
    {
        board[f] = board[s];
        board[s] = canMove;

        // check to see if checker moved to a crowning location
        if ( (f ==  0) || (f ==  1) || (f ==  2) || (f ==  3) ||
             (f == 28) || (f == 29) || (f == 30) || (f == 31)    )
        {
            board = kingMe(board, f);
        } // end if crowning
        return board;
    } // end move()

   /**
     * jump()
     * 
     * Performs a jump. Crowning is taken care of in the move() method that
     * is called upon to do the moving.
     * 
     * @param       board  - the board
     *              start  - the starting position of the moving piece
     *              jumped - the piece being jumped
     *              finish - the ending position of the moving piece
     * @return      the manipulated board 
     */
    private char[] jump(char[] board, int start, int jumped, int finish)
    {
        board         = move(board, start, finish);  // mov piece
        board[jumped] = canMove;                     // remove jumped
        return board;        
    } // end jump()

    /**
     * kingMe()
     * 
     * Given a board and a location, this method crowns the specified piece
     * a king at the location and returns the board.
     * 
     * @param       board - the board to manipulate
     *              pos   - the position in the array to crown king (0 - 31)
     * @return      the manipulated board 
     */
    private char[] kingMe(char[] board, int pos)
    {
        // RED crowning
        if (board[pos] == rChecker)     
        {
            board[pos] = rKing;
        } // end if RED
        
        // WHITE crowning
        else if (board[pos] == wChecker)  
        {
            board[pos] = wKing;
        } // end else if WHITE
        return board;
    } // end kingMe()

    /**
     * isEmpty()
     * 
     * Provided a board and a position on that board (0 - 31), this method
     * determines if that position is empty and returns that information.
     * 
     * @param       pos     - the position on the board to to test (0 - 31)
     * @return      true if pos is empty, else false 
     */
    private boolean isEmpty(char[] board, int pos)
    {
        boolean flag = false;

        if (board[pos] == canMove)
        {
            flag = true;
        } //end if empty
        return flag;
    } // end isEmpty()

    /**
     * canJump()
     * 
     * Given a board, a start location, and a location to jump, this method 
     * returns the location that can be jumped to if jumping is possible, 
     * else 99 will be returned.
     *  
     * @param       start  - starting position (0 - 31)
     *              jumped - location to jump (0 - 31)
     * @return      the location that can be jumped to (0 - 31), else 99 
     */
    private int canJump(char[] board, int start, int jumped)
    {
        int finish = 99;    // value to return if jump not possible

        // first need to see if the the pieces are opposite colors
        if ( ( (getColor(board, start) == 0) && (getColor(board, jumped) == 1) ) ||
             ( (getColor(board, start) == 1) && (getColor(board, jumped) == 0) )    )
        {
            int jCol  = checkerCol(jumped);     // jumped piece's column
            int jRow  = checkerRow(jumped);     // jumped piece's row
            int sCol  = checkerCol(start);      // starting piece's column
            int sRow  = checkerRow(start);      // starting piece's row
            int nwCol = sCol - 2;               // NW finish location column
            int nwRow = sRow - 2;               // NW finish location row
            int neCol = sCol + 2;               // NE finish location column
            int neRow = sRow - 2;               // NE finish location row
            int swCol = sCol - 2;               // SW finish location column
            int swRow = sRow + 2;               // SW finish location row
            int seCol = sCol + 2;               // SE finish location column
            int seRow = sRow + 2;               // SE finish location row
            int jumpToPos;                      // position to jump to (0 - 31)
            char piece = board[locations[sRow][sCol]]; // starting piece on the board            

            // NORTH-WEST jump location
            // pieces that can legally move NW: RED king, WHITE king, and WHITE
            // checker
            if ( (jCol < sCol) && (jRow < sRow) && (nwCol >= 0) && (nwRow >= 0) )
            {
                jumpToPos = locations[nwRow][nwCol];
                if (isEmpty(board, jumpToPos) == true)
                {
                    if ( (piece == rKing) || (piece == wKing) || (piece == wChecker) )
                    {
                        finish = locations[nwRow][nwCol];
                    } // end if legal pieces
                } // end if location empty
            } // end if NW

            // NORTH-EAST jump location
            // pieces that can legally move NW: RED king, WHITE king, and WHITE
            // checker            
            else if ( (jCol > sCol) && (jRow < sRow) && (neCol <= 7) && (neRow >= 0) )
            {
                jumpToPos = locations[neRow][neCol];
                if (isEmpty(board, jumpToPos) == true)
                {
                    if ( (piece == rKing) || (piece == wKing) || (piece == wChecker) )
                    {
                        finish = locations[neRow][neCol];
                    } // end if legal pieces
                } // end if location empty
            } // end else if NE
            
            // SOUTH-WEST jump location
            // pieces that can legally move NW: RED king, WHITE king, and RED
            // checker            
            else if ( (jCol < sCol) && (jRow > sRow) && (swCol >= 0) && (swRow <= 7) )
            {
                jumpToPos = locations[swRow][swCol];
                if (isEmpty(board, jumpToPos) == true)
                {
                    if ( (piece == rKing) || (piece == wKing) || (piece == rChecker) )
                    {
                        finish = locations[swRow][swCol];
                    } // end if legal pieces
                } // end if location empty
            } // end if SW            

            // SOUTH-EAST jump location
            // pieces that can legally move NW: RED king, WHITE king, and RED
            // checker             
            else if ( (jCol > sCol) && (jRow > sRow) && (seCol <= 7) && (seRow <= 7) )
            {
                jumpToPos = locations[seRow][seCol];
                if (isEmpty(board, jumpToPos) == true)
                {
                    if ( (piece == rKing) || (piece == wKing) || (piece == rChecker) )
                    {
                        finish = locations[seRow][seCol];
                    } // end if legal pieces
                } // end if location empty
            } // end else if SE
        } // end if opposite color pieces
        return finish;
    } // end canJump()

 /**
     * getColor()
     * 
     * Given a board and the location to look at, this method will return
     * the color of checker at that position, else 99.
     * 
     * @param       pos   -   position to examine (0 - 31)
     * @return      0 if RED, 1 if WHITE, else 99
     */
    private int getColor(char[] board, int pos)
    {
        int color = 99;  // the number of the err
        
        if (board[pos] == rChecker || board[pos] == rKing)
        {
            color = 0;   // RED
        } // end if RED piece

        if (board[pos] == wChecker || board[pos] == wKing)
        {
            color = 1;   // WHITE
        } // end if WHITE piece
        return color;
    } // end getColor()   

    /**
     * printVector()
     * 
     * Given a vector, this method print out the elements of the vector.
     * 
     * @param       vector   -   the Vector to print
     * @return      n/a
     */
    private void printVector(Vector vector)
    {
        for (int i = 0; i < vector.size(); i++)
        {
            System.out.print("[" + (((Integer)vector.get(i)).intValue()+1)
                                + "] ");
        } // end for i
        System.out.println();
    } // end printVector()
        
    //----------------------------------------------------------------------

} // end succStates class