/*
 * File: ReverseChecker.java
 * Created on 6.3.2004
 * Author: martin.pradny
 * 
 */
package ai.checkers;

import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JScrollPane;

/**
 * @author martin.pradny
 *
 */
public class ReverseChecker extends JFrame {

	//GUI declarations
	private javax.swing.JPanel jContentPane = null;

	private javax.swing.JTextArea boardPane = null;
	private javax.swing.JPanel commandPanel = null;
	private javax.swing.JLabel jLabel = null;
	private javax.swing.JTextField inputField = null;
	private javax.swing.JButton commandButton = null;
	private javax.swing.JList moveList = null;
	private javax.swing.JPanel statusPanel = null;
	private javax.swing.JLabel statusLabel = null;
	private javax.swing.JButton undoButton = null;
	private int movesTaken=0;

	//AI declarations
	private final int DEFAULTDEPTH=5;
	private State curState = new State();
	private Stack undoStates = new Stack();
	//can be unified with moves - needs correcet toString method on state
	private Vector moves = new Vector();
	private IterativeSearch search;

	public static void main(String[] args) {
		new ReverseChecker();
	}
	/**
	 * This is the default constructor
	 */
	public ReverseChecker() {
		super();
		initialize(); //GUI Initialization
		curState.initial();
		boardPane.setText(curState.toString());
		search=new IterativeSearch(curState,DEFAULTDEPTH);

	}
	/**
	 * This method initializes this
	 * 
	 * @return void
	 */
	private void initialize() {
		this.setSize(480, 215);
		this.setContentPane(getJContentPane());
		this.setVisible(true);
		this.setTitle("AI Checkers - Leviathan");
	}
	/**
	 * This method initializes jContentPane
	 * 
	 * @return javax.swing.JPanel
	 */
	private javax.swing.JPanel getJContentPane() {
		if (jContentPane == null) {
			jContentPane = new javax.swing.JPanel();
			jContentPane.setLayout(new java.awt.BorderLayout());
			jContentPane.add(getBoardPane(), java.awt.BorderLayout.CENTER);
			jContentPane.add(getCommandPanel(), java.awt.BorderLayout.SOUTH);
			JScrollPane scrollPane =
				new JScrollPane(
					getJList(),
					JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
					JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
			scrollPane.setPreferredSize(new java.awt.Dimension(150, 0));
			jContentPane.add(scrollPane, java.awt.BorderLayout.EAST);
			jContentPane.add(getStatusPanel(), java.awt.BorderLayout.NORTH);
		}
		return jContentPane;
	}
	/**
	 * This method initializes boardPane
	 * 
	 * @return javax.swing.JTextArea
	 */
	private javax.swing.JTextArea getBoardPane() {
		if (boardPane == null) {
			boardPane = new javax.swing.JTextArea();
			boardPane.setEditable(false);
			boardPane.setName("boardPane");
			boardPane.setFont(new Font("Courier New", Font.BOLD, 14));
		}
		return boardPane;
	}
	/**
	 * This method initializes commandPanel
	 * 
	 * @return javax.swing.JPanel
	 */
	private javax.swing.JPanel getCommandPanel() {
		if (commandPanel == null) {
			commandPanel = new javax.swing.JPanel();
			commandPanel.setLayout(new java.awt.BorderLayout());
			commandPanel.add(getJLabel(), java.awt.BorderLayout.WEST);
			commandPanel.add(getInputField(), java.awt.BorderLayout.CENTER);
			commandPanel.add(getCommadButton(), java.awt.BorderLayout.EAST);
			commandPanel.setName("commandPanel");
		}
		return commandPanel;
	}
	/**
	 * This method initializes jLabel
	 * 
	 * @return javax.swing.JLabel
	 */
	private javax.swing.JLabel getJLabel() {
		if (jLabel == null) {
			jLabel = new javax.swing.JLabel();
			jLabel.setText("Input:");
			jLabel.setPreferredSize(new java.awt.Dimension(50, 15));
			jLabel.setHorizontalTextPosition(javax.swing.SwingConstants.CENTER);
			jLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
		}
		return jLabel;
	}
	/**
	 * This method initializes inputField
	 * 
	 * @return javax.swing.JTextField
	 */
	private javax.swing.JTextField getInputField() {
		if (inputField == null) {
			inputField = new javax.swing.JTextField();
			inputField.setName("CommandEdit");
		}
		return inputField;
	}
	/**
	 * This method initializes commadButton
	 * 
	 * @return javax.swing.JButton
	 */
	private javax.swing.JButton getCommadButton() {
		if (commandButton == null) {
			commandButton = new javax.swing.JButton();
			commandButton.setText("Submit Command");
			commandButton.setName("CommandButton");
			commandButton.setPreferredSize(new java.awt.Dimension(150, 25));
			commandButton.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {
					String move;
					State tmpState, posState;
					boolean valid;
					Stack succStack;
					undoStates.push(curState.copy());
					if (curState.turnWhite) {
						move = "W: ";
					} else {
						move = "B: ";
					}
					
					if (inputField.getText().equals("")) {
						valid=true;
						movesTaken++;
						search.setState(curState);
						System.out.println(System.currentTimeMillis());
						//curState=search.nonrecSearch();
						curState=search.search();
						System.out.println(System.currentTimeMillis());
						move = move + curState.printMoves();
						System.out.println("");
						System.out.println(curState.data[0]);
						System.out.println(curState.data[1]);
						System.out.println(curState.data[2]);
						
						
					} else {
						tmpState=curState.copy();
						tmpState.Move(inputField.getText());
						valid=false;
						succStack=curState.genAllSuccessors();
						while (!valid && !succStack.isEmpty()) {
							posState=(State) succStack.pop();
							valid=posState.equals(tmpState);
						}
						if (valid) curState.Move(inputField.getText());
						move = move + inputField.getText();
						
					}

					
					//curState.turnWhite = !curState.turnWhite;
					if (valid) {
					moves.addElement(move);
					moveList.setListData(moves);
					updateBoard();
					
					inputField.setText("");
					updateStatus();
					} else {
						inputField.setText("Invalid move");
					}
					
				}
			});

		}
		return commandButton;
	}
	/**
	 * This method initializes jList
	 * 
	 * @return javax.swing.JList
	 */
	private javax.swing.JList getJList() {
		if (moveList == null) {
			moveList = new javax.swing.JList(moves);
			moveList.setName("MoveList");
			moveList.setEnabled(true);
			moveList.setBackground(new java.awt.Color(255, 255, 204));
		}
		return moveList;
	}
	/**
	 * This method initializes statusPanel
	 * 
	 * @return javax.swing.JPanel
	 */
	private javax.swing.JPanel getStatusPanel() {
		if (statusPanel == null) {
			statusPanel = new javax.swing.JPanel();
			statusPanel.setLayout(new java.awt.BorderLayout());
			statusPanel.add(getStatusLabel(), java.awt.BorderLayout.CENTER);
			statusPanel.add(getUndoButton(), java.awt.BorderLayout.EAST);
		}
		return statusPanel;
	}
	/**
	 * This method initializes statusLabel
	 * 
	 * @return javax.swing.JLabel
	 */
	private javax.swing.JLabel getStatusLabel() {
		if (statusLabel == null) {
			statusLabel = new javax.swing.JLabel();
			statusLabel.setText(
				"Ready to start - insert move or 'compute' command");
			statusLabel.setHorizontalAlignment(
				javax.swing.SwingConstants.CENTER);
		}
		return statusLabel;
	}
	/**
	 * This method initializes undoButton
	 * 
	 * @return javax.swing.JButton
	 */
	private javax.swing.JButton getUndoButton() {
		if (undoButton == null) {
			undoButton = new javax.swing.JButton();
			undoButton.setPreferredSize(new java.awt.Dimension(150, 11));
			undoButton.setText("Undo move");
			undoButton.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {
					if (!undoStates.isEmpty()) {
						search.removeState(curState);
						curState = (State) undoStates.pop();
						moves.remove(moves.size() - 1);
						moveList.setListData(moves);
					}
					updateBoard();
					inputField.setText("");
					updateStatus();
				}
			});
		}
		return undoButton;
	}

	private void updateStatus() {
		boolean hasJumps=!curState.genAllSuccessors().isEmpty();
		if (curState.turnWhite) {
			if (hasJumps) 
			{statusLabel.setText("White's turn");
			}else {
				statusLabel.setText("White won");
				commandButton.setEnabled(false);
				System.out.println("Moves taken "+ movesTaken);
				;} 
		} else {
			if (hasJumps) 
						{statusLabel.setText("Red's turn");
						}else {
										statusLabel.setText("Red won");
										commandButton.setEnabled(false);
							            System.out.println("Moves taken "+ movesTaken);
										;}	
		}
	}

	private void updateBoard() {
		boardPane.setText(curState.toString());
	}

	protected void processWindowEvent(WindowEvent e) {
		super.processWindowEvent(e);
		if (e.getID() == WindowEvent.WINDOW_CLOSING) {
			System.exit(0);
		}
	}
	
	

} //  @jve:visual-info  decl-index=0 visual-constraint="10,10"
